import logging
import langchain

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.recorders.instrumentation import patch_method

logger = logging.getLogger('graphsignal')


class LangChainRecorder(BaseRecorder):
    def __init__(self):
        self._library = None
        self._v1_handler = None

    def setup(self):
        if not graphsignal._tracer.auto_instrument:
            return

        self._library = signals_pb2.LibraryInfo()
        self._library.name = 'LangChain'
        version = ''
        if hasattr(langchain, '__version__') and langchain.__version__:
            parse_semver(self._library.version, langchain.__version__)
            version = langchain.__version__

        def is_v2():
            return (
                hasattr(langchain.callbacks, 'manager') and
                hasattr(langchain.callbacks.manager, 'CallbackManager') and
                hasattr(langchain.callbacks.manager, 'AsyncCallbackManager')
            )

        def is_v1():
            return hasattr(langchain.callbacks, 'get_callback_manager')


        if is_v2():
            # langchain >= 0.0.154
            from graphsignal.callbacks.langchain.v2 import GraphsignalCallbackHandler
            def after_configure(args, kwargs, ret, exc, context):
                if isinstance(ret, langchain.callbacks.manager.CallbackManager):
                    if not any(isinstance(handler, GraphsignalCallbackHandler) for handler in ret.handlers):
                        ret.add_handler(GraphsignalCallbackHandler())
                else:
                    logger.error(f'Cannot add callback for LangChain {version}')
            if not patch_method(langchain.callbacks.manager.CallbackManager, 'configure', after_func=after_configure):
                logger.error(f'Cannot instrument LangChain {version}')

            from graphsignal.callbacks.langchain.v2 import GraphsignalAsyncCallbackHandler
            def after_async_configure(args, kwargs, ret, exc, context):
                if isinstance(ret, langchain.callbacks.manager.AsyncCallbackManager):
                    if not any(isinstance(handler, GraphsignalAsyncCallbackHandler) for handler in ret.handlers):
                        ret.add_handler(GraphsignalAsyncCallbackHandler())
                else:
                    logger.error(f'Cannot add callback for LangChain {version}')
            if not patch_method(langchain.callbacks.manager.AsyncCallbackManager, 'configure', after_func=after_async_configure):
                logger.error(f'Cannot instrument LangChain {version}')

        elif is_v1():
            # compatibility with langchain <= 0.0.153
            from graphsignal.callbacks.langchain.v1 import GraphsignalCallbackHandler
            self._v1_handler = GraphsignalCallbackHandler()
            langchain.callbacks.get_callback_manager().add_handler(self._v1_handler)

        else:
            logger.error(f'Cannot auto-instrument LangChain {version}')

    def shutdown(self):
        if self._v1_handler:
            langchain.callbacks.get_callback_manager().remove_handler(self._v1_handler)
            self._v1_handler = None

    def on_span_read(self, proto, context, options):
        if self._library:
            proto.libraries.append(self._library)



from typing import Union, Any, Optional, Dict
import logging
import sys
import time
import traceback
import contextvars
import uuid
import hashlib

import graphsignal
from graphsignal import version
from graphsignal.proto import signals_pb2
from graphsignal.data import compute_data_stats, encode_data_sample
from graphsignal.proto_utils import parse_semver

logger = logging.getLogger('graphsignal')


span_stack_var = contextvars.ContextVar('span_stack_var', default=[])


def clear_span_stack():
    span_stack_var.set([])


def push_current_span(span):
    span_stack_var.set(span_stack_var.get() + [span])


def pop_current_span(span):
    span_stack = span_stack_var.get()
    if len(span_stack) > 0 and span_stack[-1] == span:
        span_stack_var.set(span_stack[:-1])
        return span_stack[-1]
    return None


def get_root_span():
    span_stack = span_stack_var.get()
    if len(span_stack) > 0:
        return span_stack[0]
    return None


def get_parent_span():
    span_stack = span_stack_var.get()
    if len(span_stack) > 1:
        return span_stack[-2]
    return None


def get_current_span():
    span_stack = span_stack_var.get()
    if len(span_stack) > 0:
        return span_stack[-1]
    return None


def _tracer():
    return graphsignal._tracer


class TraceOptions:
    __slots__ = [
        'record_samples',
        'record_metrics',
        'enable_profiling'
    ]

    def __init__(self, 
            record_samples: bool = True,
            record_metrics: bool = True,
            enable_profiling: bool = False):
        self.record_samples = record_samples
        self.record_metrics = record_metrics
        self.enable_profiling = enable_profiling

DEFAULT_OPTIONS = TraceOptions()


class DataObject:
    __slots__ = [
        'name',
        'obj',
        'counts',
        'record_data_sample',
        'check_missing_values'
    ]

    def __init__(self, name, obj, counts=None, record_data_sample=True, check_missing_values=False):
        self.name = name
        self.obj = obj
        self.counts = counts
        self.record_data_sample = record_data_sample
        self.check_missing_values = check_missing_values


class Span:
    MAX_RUN_TAGS = 10
    MAX_SPAN_TAGS = 10
    MAX_RUN_PARAMS = 10
    MAX_TRACE_PARAMS = 10
    MAX_DATA_OBJECTS = 10
    MAX_SAMPLE_BYTES = 256 * 1024

    __slots__ = [
        '_operation',
        '_tags',
        '_params',
        '_options',
        '_root_span',
        '_parent_span',
        '_is_sampling',
        '_is_root',
        '_latency_ns',
        '_context',
        '_proto',
        '_is_started',
        '_is_stopped',
        '_start_counter',
        '_stop_counter',
        '_exc_infos',
        '_data_objects',
        '_is_latency_outlier',
        '_has_missing_values'
    ]

    def __init__(self, operation, tags=None, options=None):
        push_current_span(self)

        self._is_started = False

        if not operation:
            logger.error('operation is required')
            return
        if tags is not None:
            if not isinstance(tags, dict):
                logger.error('tags must be dict')
                return
            if len(tags) > Span.MAX_SPAN_TAGS:
                logger.error('too many tags (>{0})'.format(Span.MAX_SPAN_TAGS))
                return

        self._operation = _sanitize_str(operation)
        if tags is not None:
            self._tags = dict(tags)
        else:
            self._tags = None
        self._params = None
        if options is None:
            self._options = DEFAULT_OPTIONS
        else:
            self._options = options
        self._is_stopped = False
        self._root_span = None
        self._parent_span = None
        self._is_sampling = False
        self._is_root = False
        self._latency_ns = None
        self._start_counter = None
        self._stop_counter = None
        self._context = False
        self._proto = None
        self._exc_infos = None
        self._data_objects = None
        self._is_latency_outlier = False
        self._has_missing_values = False

        try:
            self._start()
        except Exception:
            logger.error('Error starting span', exc_info=True)
            self._is_stopped = True

    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(self, *exc_info):
        if exc_info and exc_info[1] and isinstance(exc_info[1], Exception):
            if not self._exc_infos:
                self._exc_infos = []
            self._exc_infos.append(exc_info)
        self.stop()
        return False

    async def __aexit__(self, *exc_info):
        if exc_info and exc_info[1] and isinstance(exc_info[1], Exception):
            if not self._exc_infos:
                self._exc_infos = []
            self._exc_infos.append(exc_info)
        self.stop()
        return False

    def _init_sampling(self, sampling_type):
        self._is_sampling = True
        self._proto = signals_pb2.Span()
        self._proto.span_id = _uuid_sha1(size=12)
        self._proto.sampling_type = sampling_type
        self._proto.tracer_info.tracer_type = signals_pb2.TracerInfo.TracerType.PYTHON_TRACER
        parse_semver(self._proto.tracer_info.version, version.__version__)
        self._context = {}

    def _propagate_sampling(self, level=0):
        if level > 10:
            return
        if self._parent_span:
            if not self._parent_span.is_sampling():
                self._parent_span._init_sampling(sampling_type=signals_pb2.Span.SamplingType.CHILD_SAMPLING)
                self._parent_span._propagate_sampling(level+1)

    def _start(self):
        if self._is_started:
            return
        if self._is_stopped:
            return

        if _tracer().debug_mode:
            logger.debug(f'Starting span {self._operation}')

        self._parent_span = get_parent_span()
        self._is_root = not self._parent_span
        self._root_span = self._parent_span.root_span() if self._parent_span else self

        if self._options.record_samples:
            # sample if parent span is being sampled or this span is root
            if self._is_root:
                if _tracer().random_sampler(self._operation).sample('random'):
                    self._init_sampling(sampling_type=signals_pb2.Span.SamplingType.RANDOM_SAMPLING)
            else:
                if self._parent_span.is_sampling():
                    self._init_sampling(sampling_type=signals_pb2.Span.SamplingType.PARENT_SAMPLING)

            # emit start event
            if self._is_sampling:
                try:
                    _tracer().emit_span_start(self._proto, self._context, self._options)
                except Exception as exc:
                    logger.error('Error in span start event handlers', exc_info=True)

        self._start_counter = time.perf_counter_ns()
        self._is_started = True

    def _measure(self) -> None:
        self._stop_counter = time.perf_counter_ns()

    def _stop(self) -> None:
        if not self._is_started:
            return
        if self._is_stopped:
            return
        self._is_stopped = True

        if _tracer().debug_mode:
            logger.debug(f'Stopping span {self._operation}')

        if self._stop_counter is None:
            self._measure()
        self._latency_ns = self._stop_counter - self._start_counter

        now = time.time()
        end_us = int(now * 1e6)
        start_us = int(end_us - self._latency_ns / 1e3)
        now = int(now)

        # emit stop event
        if self._is_sampling:
            try:
                _tracer().emit_span_stop(self._proto, self._context, self._options)
            except Exception as exc:
                logger.error('Error in span stop event handlers', exc_info=True)

        span_tags = self._span_tags()

        # if exception, but the span is not being recorded, try to start tracing
        if self._options.record_samples:
            if not self._is_sampling and self._exc_infos and len(self._exc_infos) > 0:
                if _tracer().random_sampler(self._operation).sample('exceptions'):
                    self._init_sampling(sampling_type=signals_pb2.Span.SamplingType.ERROR_SAMPLING)
                    self._propagate_sampling()

        # check for outliers
        lo_sampler = _tracer().lo_sampler(self._operation)
        if self._options.record_samples:
            self._is_latency_outlier = lo_sampler.sample(self._latency_ns / 1e9)
            if not self._is_sampling and self._is_latency_outlier:
                if _tracer().random_sampler(self._operation).sample('latency-outliers'):
                    self._init_sampling(sampling_type=signals_pb2.Span.SamplingType.ERROR_SAMPLING)
                    self._propagate_sampling()
        lo_sampler.update(self._latency_ns / 1e9)

        # update RED metrics
        if self._options.record_metrics:
            _tracer().metric_store().update_histogram(
                scope='performance', name='latency', tags=span_tags, value=self._latency_ns, update_ts=now, is_time=True)
            _tracer().metric_store().inc_counter(
                scope='performance', name='call_count', tags=span_tags, value=1, update_ts=now)
            if self._exc_infos and len(self._exc_infos) > 0:
                for exc_info in self._exc_infos:
                    if exc_info[0] is not None:
                        _tracer().metric_store().inc_counter(
                            scope='performance', name='exception_count', tags=span_tags, value=1, update_ts=now)
                        self.set_tag('exception', exc_info[0].__name__)

        # compute data statistics
        data_stats = None
        if self._options.record_metrics and self._data_objects is not None:
            data_stats = {}
            for data_obj in self._data_objects.values():
                try:
                    stats = compute_data_stats(data_obj.obj)
                    if not stats:
                        continue
                    data_stats[data_obj.name] = stats

                    # add/update extra counts to computed counts
                    if data_obj.counts is not None:
                        stats.counts.update(data_obj.counts)

                    # check missing values
                    if data_obj.check_missing_values:
                        if _tracer().mv_sampler(self._operation).sample(data_obj.name, stats.counts):
                            self._has_missing_values = True

                    # update data metrics
                    data_tags = span_tags.copy()
                    data_tags['data'] = data_obj.name
                    for count_name, count in stats.counts.items():
                        _tracer().metric_store().inc_counter(
                            scope='data', name=count_name, tags=data_tags, value=count, update_ts=now)
                except Exception as exc:
                    logger.error('Error computing data stats', exc_info=True)

        # if missing values detected, but the span is not being recorded, try to start sampling
        if self._options.record_samples:
            if not self._is_sampling and self._has_missing_values:
                if _tracer().random_sampler(self._operation).sample('missing-values'):
                    self._init_sampling(sampling_type=signals_pb2.Span.SamplingType.ERROR_SAMPLING)
                    self._propagate_sampling()

        # emit read event
        if self._is_sampling:
            try:
                _tracer().emit_span_read(self._proto, self._context, self._options)
            except Exception as exc:
                logger.error('Error in span read event handlers', exc_info=True)

        # update recorder metrics
        if self._options.record_metrics and _tracer().check_metric_read_interval(now):
            try:
                _tracer().emit_metric_update()
                _tracer().set_metric_read(now)
            except Exception as exc:
                logger.error('Error in span read event handlers', exc_info=True)

        # fill and upload span
        if self._is_sampling:
            # copy data to span proto
            self._proto.start_us = start_us
            self._proto.end_us = end_us
            if self._is_root:
                self._proto.labels.insert(0, 'root')
            if self._exc_infos and len(self._exc_infos) > 0:
                self._proto.labels.append('exception')
            if self._is_latency_outlier:
                self._proto.labels.append('latency-outlier')
            if self._has_missing_values:
                self._proto.labels.append('missing-values')
            self._proto.process_usage.start_ms = _tracer()._process_start_ms

            # copy span context
            self._proto.context.start_ns = self._start_counter
            self._proto.context.end_ns = self._stop_counter
            if self._parent_span and self._parent_span._proto:
                self._proto.context.parent_span_id = self._parent_span._proto.span_id
            if self._root_span and self._root_span._proto:
                self._proto.context.root_span_id = self._root_span._proto.span_id

            # copy tags
            for key, value in span_tags.items():
                tag = self._proto.tags.add()
                tag.key = _sanitize_str(key)
                tag.value = _sanitize_str(value)

            # copy params
            if self._params is not None:
                for name, value in self._params.items():
                    param = self._proto.params.add()
                    param.name = _sanitize_str(name)
                    param.value = _sanitize_str(value)

            # copy exception
            if self._exc_infos:
                for exc_info in self._exc_infos:
                    exception_proto = self._proto.exceptions.add()
                    if exc_info[0] and hasattr(exc_info[0], '__name__'):
                        exception_proto.exc_type = str(exc_info[0].__name__)
                    if exc_info[1]:
                        exception_proto.message = str(exc_info[1])
                    if exc_info[2]:
                        frames = traceback.format_tb(exc_info[2])
                        if len(frames) > 0:
                            exception_proto.stack_trace = ''.join(frames)

            # copy data stats
            if data_stats is not None:
                for data_name, stats in data_stats.items():
                    data_stats_proto = self._proto.data_profile.add()
                    data_stats_proto.data_name = data_name
                    if stats.type_name:
                        data_stats_proto.data_type = stats.type_name
                    if stats.shape:
                        data_stats_proto.shape[:] = stats.shape
                    for counter_name, count in stats.counts.items():
                        if count > 0:
                            dc = data_stats_proto.counts.add()
                            dc.name = counter_name
                            dc.count = count

                    # copy data sample
                    data_obj = self._data_objects[data_name]
                    if _tracer().record_data_samples and data_obj.record_data_sample:
                        try:
                            sample = encode_data_sample(data_obj.obj)
                            if sample is not None and len(sample.content_bytes) <= Span.MAX_SAMPLE_BYTES:
                                sample_proto = self._proto.data_samples.add()
                                sample_proto.data_name = data_name
                                sample_proto.content_type = sample.content_type
                                sample_proto.content_bytes = sample.content_bytes
                        except Exception as exc:
                            logger.debug('Error encoding {0} sample for operation {1}'.format(data_name, self._operation))

            # queue span proto for upload
            _tracer().uploader().upload_span(self._proto)

        # trigger upload
        if self._is_root:
            _tracer().tick(now)

    def measure(self) -> None:
        if not self._is_stopped:
            self._measure()

    def stop(self) -> None:
        try:
            self._stop()
        except Exception:
            logger.error('Error stopping span', exc_info=True)
        finally:
            self._is_stopped = True
            pop_current_span(self)

    def root_span(self) -> Optional['Span']:
        return self._root_span

    def parent_span(self) -> Optional['Span']:
        return self._parent_span

    def is_sampling(self):
        return self._is_sampling

    def set_tag(self, key: str, value: str) -> None:
        if not key:
            logger.error('set_tag: key must be provided')
            return
        if value is None:
            logger.error('set_tag: value must be provided')
            return

        if self._tags is None:
            self._tags = {}

        if len(self._tags) > Span.MAX_SPAN_TAGS:
            logger.error('set_tag: too many tags (>{0})'.format(Span.MAX_SPAN_TAGS))
            return

        self._tags[key] = value

    def set_param(self, name: str, value: str) -> None:
        if not name:
            logger.error('set_param: name must be provided')
            return

        if self._params is None:
            self._params = {}

        if len(self._params) > Span.MAX_TRACE_PARAMS:
            logger.error('set_param: too many params (>{0})'.format(Span.MAX_TRACE_PARAMS))
            return

        self._params[name] = value

    def add_exception(self, exc: Optional[Exception] = None, exc_info: Optional[bool] = None) -> None:
        if exc is not None and not isinstance(exc, Exception):
            logger.error('add_exception: exc must be instance of Exception')
            return

        if exc_info is not None and not isinstance(exc_info, bool):
            logger.error('add_exception: exc_info must be bool')
            return

        if self._exc_infos is None:
            self._exc_infos = []

        if exc:
            self._exc_infos.append((exc.__class__, str(exc), exc.__traceback__))
        elif exc_info == True:
            self._exc_infos.append(sys.exc_info())

    def set_data(self, 
            name: str, 
            obj: Any, 
            counts: Optional[Dict[str, int]] = None, 
            record_data_sample: Optional[bool] = True, 
            check_missing_values: Optional[bool] = False) -> None:
        if self._data_objects is None:
            self._data_objects = {}

        if name and not isinstance(name, str):
            logger.error('set_data: name must be string')
            return

        if counts and not isinstance(counts, dict):
            logger.error('append_data: name must be dict')
            return

        if len(self._data_objects) > Span.MAX_DATA_OBJECTS:
            logger.error('set_data: too many data objects (>{0})'.format(Span.MAX_DATA_OBJECTS))
            return

        self._data_objects[name] = DataObject(
            name=name,
            obj=obj,
            counts=counts,
            record_data_sample=record_data_sample,
            check_missing_values=check_missing_values)

    def append_data(self, 
            name: str, 
            obj: Any, 
            counts: Optional[Dict[str, int]] = None, 
            record_data_sample: Optional[bool] = True, 
            check_missing_values: Optional[bool] = False) -> None:
        if self._data_objects is None:
            self._data_objects = {}

        if name and not isinstance(name, str):
            logger.error('append_data: name must be string')
            return

        if counts and not isinstance(counts, dict):
            logger.error('append_data: name must be dict')
            return

        if len(self._data_objects) > Span.MAX_DATA_OBJECTS:
            logger.error('append_data: too many data objects (>{0})'.format(Span.MAX_DATA_OBJECTS))
            return

        if name in self._data_objects:
            data_obj = self._data_objects[name]
            data_obj.obj += obj
            if counts:
                if data_obj.counts is None:
                    data_obj.counts = {}
                for count_name, value in counts.items():
                    if count_name in data_obj.counts:
                        data_obj.counts[count_name] += value
                    else:
                        data_obj.counts[count_name] = value
        else:
            self._data_objects[name] = DataObject(
                name=name,
                obj=obj,
                counts=counts,
                record_data_sample=record_data_sample,
                check_missing_values=check_missing_values)

    def _span_tags(self, extra_tags=None):
        span_tags = {
            'deployment': _tracer().deployment, 
            'operation': self._operation}
        if _tracer().hostname:
            span_tags['hostname'] = _tracer().hostname
        if _tracer().tags is not None:
            span_tags.update(_tracer().tags)
        context_tags = _tracer().context_tags.get()
        if len(context_tags) > 0:
            span_tags.update(context_tags)
        if self._tags is not None:
            span_tags.update(self._tags)
        if extra_tags is not None:
            span_tags.update(extra_tags)
        return span_tags

    def repr(self):
        return 'Span({0})'.format(self._operation)


def _sanitize_str(val, max_len=250):
    if not isinstance(val, str):
        return str(val)[:max_len]
    else:
        return val[:max_len]


def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]


def _uuid_sha1(size=-1):
    return _sha1(str(uuid.uuid4()), size)



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import openai
import transformers
from transformers.tools import OpenAiAgent

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.spans import DEFAULT_OPTIONS
from graphsignal.recorders.openai_recorder import OpenAIRecorder
from graphsignal.recorders.huggingface_recorder import HuggingFaceRecorder
from graphsignal.proto_utils import find_tag, find_param, find_data_sample

logger = logging.getLogger('graphsignal')


class HuggingFaceRecorderTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    async def test_record(self):
        recorder = HuggingFaceRecorder()
        recorder.setup()
        proto = signals_pb2.Span()
        context = {}
        recorder.on_span_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_span_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_span_read(proto, context, DEFAULT_OPTIONS)

        self.assertEqual(proto.libraries[0].name, 'Transformers')

    @patch.object(Uploader, 'upload_span')
    @patch.object(openai.Completion, 'create')
    @patch.object(transformers.tools.agents, 'get_remote_tools')
    @patch.object(transformers.tools.RemoteTool, '__call__')
    async def test_trace_run(self, mocked_remote_tool, mocked_get_remote_tools, mocked_create, mocked_upload_span):
        now = int(time.time())

        transformers.tools.agents.HUGGINGFACE_DEFAULT_TOOLS_FROM_HUB = []
        from transformers.tools.agents import PreTool
        mocked_get_remote_tools.return_value = {
            'translator': PreTool(
                task='translation', 
                description="This is a tool that translates text from a language to another. It takes three inputs: `text`, which should be the text to translate, `src_lang`, which should be the language of the text to translate and `tgt_lang`, which should be the language for the desired ouput language. Both `src_lang` and `tgt_lang` are written in plain English, such as 'Romanian', or 'Albanian'. It returns the text translated in `tgt_lang`.", repo_id=None)
        }

        mocked_remote_tool.return_value = 'Bonjour, comment allez-vous?'

        mocked_create.return_value = {
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "logprobs": None,
                    "text": " tool: `translator` to translate the text to French.\n\nAnswer:\n```py\ntranslated_text = translator(text=\"Hi, how are you?\", src_lang=\"English\", tgt_lang=\"French\")\nprint(f\"The translated text is {translated_text}.\")\n```"
                },
            ],
            "created": now,
            "id": "cmpl-id",
            "model": "text-davinci-003",
            "object": "text_completion",
            "usage": {
                "completion_tokens": 69,
                "prompt_tokens": 1595,
                "total_tokens": 1664
            }
        }

        # mocking overrides autoinstrumentation, reinstrument
        oai_rec = OpenAIRecorder()
        oai_rec.setup()
        hf_rec = HuggingFaceRecorder()
        hf_rec.setup()

        agent = OpenAiAgent(model="text-davinci-003")
        agent.run('Translate the following text to French: System update available.', remote=True)

        hf_rec.shutdown()
        oai_rec.shutdown()

        s3 = mocked_upload_span.call_args_list[0][0][0]
        s2 = mocked_upload_span.call_args_list[1][0][0]
        s1 = mocked_upload_span.call_args_list[2][0][0]

        self.assertEqual(find_tag(s1, 'component'), 'Agent')
        self.assertEqual(find_tag(s1, 'operation'), 'transformers.tools.Agent.run')
        self.assertIsNotNone(find_data_sample(s1, 'task'))
        self.assertIsNotNone(find_data_sample(s1, 'output'))

        self.assertEqual(find_tag(s2, 'component'), 'Tool')
        self.assertEqual(find_tag(s2, 'operation'), 'transformers.tools.TranslationTool')
        self.assertIsNotNone(find_tag(s2, 'endpoint'))
        self.assertIsNotNone(find_data_sample(s2, 'inputs'))
        self.assertIsNotNone(find_data_sample(s2, 'outputs'))

        self.assertEqual(find_tag(s3, 'component'), 'LLM')



import logging
import sys
import time
import chromadb

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import patch_method, unpatch_method, instrument_method, uninstrument_method, read_args
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

class ChromaRecorder(BaseRecorder):
    def __init__(self):
        self._library = None

    def setup(self):
        if not graphsignal._tracer.auto_instrument:
            return

        self._library = signals_pb2.LibraryInfo()
        self._library.name = 'Chroma'
        parse_semver(self._library.version, chromadb.__version__)

        from chromadb.api.models.Collection import Collection
        for method_name in ['add', 'update', 'upsert']:
            instrument_method(Collection, method_name, f'chroma.collection.{method_name}', trace_func=self.trace_add)
        instrument_method(Collection, 'delete', 'chroma.collection.delete', trace_func=self.trace_delete)
        instrument_method(Collection, 'get', 'chroma.collection.get', trace_func=self.trace_get)
        instrument_method(Collection, 'query', 'chroma.collection.query', trace_func=self.trace_query)

    def shutdown(self):
        from chromadb.api.models.Collection import Collection
        for method_name in ['add', 'update', 'upsert']:
            uninstrument_method(Collection, method_name)
        uninstrument_method(Collection, 'delete')
        uninstrument_method(Collection, 'get')
        uninstrument_method(Collection, 'query')

    def _fill(self, span, collection):
        span.set_tag('component', 'Memory')
        span.set_tag('index', collection.name)
        span.set_param('collection', collection.name)

    def trace_add(self, span, args, kwargs, ret, exc):
        params = read_args(args, kwargs, [
            'collection',
            'ids',
            'embeddings', 
            'metadatas', 
            'documents',
            'increment_index'])

        self._fill(span, params['collection'])

        if 'increment_index' in params:
            span.set_param('increment_index', params['increment_index'])

        if 'documents' in params:
            span.set_data('documents', params['documents'])

        if 'embeddings' in params:
            span.set_data('embeddings', params['embeddings'], record_data_sample=False)

    def trace_delete(self, span, args, kwargs, ret, exc):
        params = read_args(args, kwargs, [
            'collection',
            'ids',
            'where',
            'where_document'])

        self._fill(span, params['collection'])

    def trace_get(self, span, args, kwargs, ret, exc):
        params = read_args(args, kwargs, [
            'collection',
            'ids',
            'where', 
            'limit', 
            'offset', 
            'where_document',
            'include'])

        self._fill(span, params['collection'])

        if ret:
            if 'embeddings' in ret:
                span.set_data('embeddings', ret['embeddings'], record_data_sample=False)
            if 'documents' in ret:
                span.set_data('documents', ret['documents'])

    def trace_query(self, span, args, kwargs, ret, exc):
        params = read_args(args, kwargs, [
            'collection',
            'query_embeddings',
            'query_texts', 
            'n_results', 
            'documents', 
            'where',
            'where_document',
            'include'])

        self._fill(span, params['collection'])

        if 'n_results' in params:
            span.set_param('n_results', params['n_results'])

        if 'query_embeddings' in params:
            span.set_data('query_embeddings', params['query_embeddings'], record_data_sample=False)

        if 'query_texts' in params:
            span.set_data('query_texts', params['query_texts'])

        if ret:
            if 'embeddings' in ret:
                span.set_data('embeddings', ret['embeddings'])
            if 'documents' in ret:
                span.set_data('documents', ret['documents'])
            if 'distances' in ret:
                span.set_data('distances', ret['distances'])

    def on_span_read(self, proto, context, options):
        if self._library:
            proto.libraries.append(self._library)



import logging
from typing import Any, Dict, List, Optional, Union
from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, LLMResult, ChatResult, BaseMessage
from uuid import UUID

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.spans import get_current_span, push_current_span, clear_span_stack

logger = logging.getLogger('graphsignal')

# prevent memory leak, if spans are not stopped for some reason
MAX_TRACES = 10000 

# is thread-safe, because run_id is unique per thread
_span_map = {}


class GraphsignalCallbackHandler(BaseCallbackHandler):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._context_tags = graphsignal._tracer.context_tags.get().copy()
        self._parent_span = get_current_span()

    def _start_trace(self, parent_run_id, run_id, operation):
        if run_id in _span_map or len(_span_map) > MAX_TRACES:
            return None

        # do not rely on contextvars in callbacks
        clear_span_stack()

        # set parent span
        if not parent_run_id:
            # initialize handler
            if self._parent_span:
                push_current_span(self._parent_span)
        else:
            parent_span = _span_map.get(parent_run_id)
            if parent_span:
                push_current_span(parent_span)
            else:
                return None

        # set context tags
        if self._context_tags:
            for key, value in self._context_tags.items():
                graphsignal.set_context_tag(key, value)

        span = graphsignal.start_trace(operation)
        _span_map[run_id] = span
        return span

    def _current_span(self, run_id):
        return _span_map.get(run_id)

    def _stop_trace(self, run_id):
        span = _span_map.pop(run_id, None)
        if span:
            span.stop()

    def on_llm_start(
            self, 
            serialized: Dict[str, Any], 
            prompts: List[str], 
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> None:
        try:
            operation = 'langchain.llms.' + serialized.get('name', 'LLM')
            span = self._start_trace(parent_run_id, run_id, operation)
            if span:
                span.set_tag('component', 'LLM')
                if prompts:
                    span.set_data('prompts', prompts)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_chat_model_start(
            self,
            serialized: Dict[str, Any],
            messages: List[List[BaseMessage]],
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> Any:
        try:
            operation = 'langchain.llms.' + serialized.get('name', 'LLM')
            span = self._start_trace(parent_run_id, run_id, operation)
            if span:
                span.set_tag('component', 'LLM')
                if messages:
                    span.set_data('messages', messages)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    
    def on_llm_new_token(
            self,
            token: str,
            **kwargs: Any) -> None:
        pass

    def on_llm_end(
            self, 
            response: LLMResult, 
            *, 
            run_id: UUID,
            **kwargs: Any) -> None:
        try:
            span = self._current_span(run_id)
            if span:
                if isinstance(response, (LLMResult, ChatResult)) and hasattr(response, 'llm_output'):
                    span.set_data('output', response.llm_output)
                    span.set_data('generations', response.generations)
            self._stop_trace(run_id)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_llm_error(
            self, 
            error: Union[Exception, KeyboardInterrupt], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        try:
            span = self._current_span(run_id)
            if span:
                if isinstance(error, Exception):
                    span.add_exception(error)
            self._stop_trace(run_id)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_chain_start(
            self, 
            serialized: Dict[str, Any], 
            inputs: Dict[str, Any], 
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> None:
        try:
            chain_name = serialized.get('name', 'Chain')
            operation = 'langchain.chains.' + chain_name
            span = self._start_trace(parent_run_id, run_id, operation)
            if span:
                span.set_tag('component', 'Agent')
                if inputs:
                    span.set_data('inputs', inputs)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_chain_end(
            self, 
            outputs: Dict[str, Any], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        try:
            span = self._current_span(run_id)
            if span:
                if outputs:
                    span.set_data('outputs', outputs)
            self._stop_trace(run_id)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_chain_error(
            self, 
            error: Union[Exception, KeyboardInterrupt], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        try:
            span = self._current_span(run_id)
            if span:
                if isinstance(error, Exception):
                    span.add_exception(error)
            self._stop_trace(run_id)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_tool_start(
            self, 
            serialized: Dict[str, Any], 
            input_str: str, 
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> None:
        try:
            operation = 'langchain.agents.tools.' + serialized.get('name', 'Tool')
            span = self._start_trace(parent_run_id, run_id, operation)
            if span:
                span.set_tag('component', 'Tool')
                if input_str:
                    span.set_data('input', input_str)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_tool_end(
            self, 
            output: str, 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        try:
            span = self._current_span(run_id)
            if span:
                if output:
                    span.set_data('output', output)
            self._stop_trace(run_id)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_tool_error(
            self, 
            error: Union[Exception, KeyboardInterrupt], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        try:
            span = self._current_span(run_id)
            if span:
                if isinstance(error, Exception):
                    span.add_exception(error)
            self._stop_trace(run_id)
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_agent_action(
            self,
            action: AgentAction,
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,            
            **kwargs: Any) -> None:
        try:
            span = self._current_span(run_id)
            if span:
                if isinstance(action, AgentAction):
                    span.append_data('actions', [obj_to_dict(action)])
        except Exception:
            logger.error('Error in LangChain callback handler', exc_info=True)

    def on_agent_finish(
            self,
            finish: AgentFinish,
            **kwargs: Any) -> None:
        pass

    def on_text(
            self,
            text: str,
            **kwargs: Any) -> Any:
        pass


class GraphsignalAsyncCallbackHandler(GraphsignalCallbackHandler):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    async def on_llm_start(
            self, 
            serialized: Dict[str, Any], 
            prompts: List[str], 
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> None:
        super().on_llm_start(
            serialized, 
            prompts, 
            run_id=run_id, 
            parent_run_id=parent_run_id, 
            **kwargs)

    async def on_chat_model_start(
            self,
            serialized: Dict[str, Any],
            messages: List[List[BaseMessage]],
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> Any:
        super().on_chat_model_start(
            serialized, 
            messages, 
            run_id=run_id, 
            parent_run_id=parent_run_id, 
            **kwargs)

    async def on_llm_new_token(
            self,
            token: str,
            **kwargs: Any) -> None:
        pass

    async def on_llm_end(
            self, 
            response: LLMResult, 
            *, 
            run_id: UUID,
            **kwargs: Any) -> None:
        super().on_llm_end(
            response,
            run_id=run_id,
            **kwargs)

    async def on_llm_error(
            self, 
            error: Union[Exception, KeyboardInterrupt], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        super().on_llm_error(
            error,
            run_id=run_id,
            **kwargs)

    async def on_chain_start(
            self, 
            serialized: Dict[str, Any], 
            inputs: Dict[str, Any], 
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> None:
        super().on_chain_start(
            serialized,
            inputs,
            run_id=run_id,
            parent_run_id=parent_run_id,
            **kwargs)

    async def on_chain_end(
            self, 
            outputs: Dict[str, Any], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        super().on_chain_end(
            outputs,
            run_id=run_id,
            **kwargs)

    async def on_chain_error(
            self, 
            error: Union[Exception, KeyboardInterrupt], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        super().on_chain_error(
            error,
            run_id=run_id,
            **kwargs)

    async def on_tool_start(
            self, 
            serialized: Dict[str, Any], 
            input_str: str, 
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any) -> None:
        super().on_tool_start(
            serialized,
            input_str,
            run_id=run_id,
            parent_run_id=parent_run_id,
            **kwargs)

    async def on_tool_end(
            self, 
            output: str, 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        super().on_tool_end(
            output,
            run_id=run_id,
            **kwargs)

    async def on_tool_error(
            self, 
            error: Union[Exception, KeyboardInterrupt], 
            *,
            run_id: UUID,
            **kwargs: Any) -> None:
        super().on_tool_error(
            error,
            run_id=run_id,
            **kwargs)

    async def on_agent_action(
            self,
            action: AgentAction,
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,            
            **kwargs: Any) -> None:
        pass

    async def on_agent_finish(
            self,
            finish: AgentFinish,
            **kwargs: Any) -> None:
        pass

    async def on_text(
            self,
            text: str,
            **kwargs: Any) -> Any:
        pass



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
import random

import graphsignal
from graphsignal.samplers.latency_outliers import LatencyOutlierSampler

logger = logging.getLogger('graphsignal')

class LatencyOutlierSamplerTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_sample(self):
        sampler = LatencyOutlierSampler()

        for i in range(500):
            val = random.randint(100, 200)
            self.assertFalse(sampler.sample(val))
            sampler.update(val)
        self.assertTrue(sampler.sample(1000))


import logging
import sys
import time
import importlib
import socket
import contextvars
import importlib.abc

from graphsignal.uploader import Uploader
from graphsignal.metrics import MetricStore
from graphsignal.logs import LogStore
from graphsignal.samplers.random_samples import RandomSampler
from graphsignal.samplers.latency_outliers import LatencyOutlierSampler
from graphsignal.samplers.missing_values import MissingValueSampler

logger = logging.getLogger('graphsignal')


class GraphsignalTracerLogHandler(logging.Handler):
    def __init__(self, tracer):
        super().__init__()
        self._tracer = tracer  

    def emit(self, record):
        try:
            log_tags = {'deployment': self._tracer.deployment}
            if self._tracer.hostname:
                log_tags['hostname'] = self._tracer.hostname
            if self._tracer.tags is not None:
                log_tags.update(self._tracer.tags)

            exception = None
            if record.exc_info and isinstance(record.exc_info, tuple):
                exception = self.format(record)

            self._tracer.log_store().log_tracer_message(
                tags=log_tags,
                level=record.levelname,
                message=record.getMessage(),
                exception=exception)
        except Exception:
            pass


RECORDER_SPECS = {
    '(default)': [
        ('graphsignal.recorders.cprofile_recorder', 'CProfileRecorder'),
        ('graphsignal.recorders.process_recorder', 'ProcessRecorder'),
        ('graphsignal.recorders.nvml_recorder', 'NVMLRecorder')],
    'torch': [
        ('graphsignal.recorders.pytorch_recorder', 'PyTorchRecorder'),
        ('graphsignal.recorders.kineto_recorder', 'KinetoRecorder')],
    'yappi': [('graphsignal.recorders.yappi_recorder', 'YappiRecorder')],
    'openai': [('graphsignal.recorders.openai_recorder', 'OpenAIRecorder')],
    'langchain': [('graphsignal.recorders.langchain_recorder', 'LangChainRecorder')],
    'llama_index': [('graphsignal.recorders.llama_index_recorder', 'LlamaIndexRecorder')],
    'transformers': [('graphsignal.recorders.huggingface_recorder', 'HuggingFaceRecorder')],
    'banana_dev': [('graphsignal.recorders.banana_recorder', 'BananaRecorder')],
    'chromadb': [('graphsignal.recorders.chroma_recorder', 'ChromaRecorder')],
    'autogpt.prompt': [('graphsignal.recorders.autogpt_recorder', 'AutoGPTRecorder')]
}

class SourceLoaderWrapper(importlib.abc.SourceLoader):
    def __init__(self, loader, tracer):
        self._loader = loader
        self._tracer = tracer

    def create_module(self, spec):
        return self._loader.create_module(spec)

    def exec_module(self, module):
        self._loader.exec_module(module)
        try:
            self._tracer.initialize_recorders_for_module(module.__name__)
        except Exception:
            logger.error('Error initializing recorders for module %s', module.__name__, exc_info=True)

    def load_module(self, fullname):
        self._loader.load_module(fullname)

    def get_data(self, path):
        return self._loader.get_data(path)

    def get_filename(self, fullname):
        return self._loader.get_filename(fullname)


class SupportedModuleFinder(importlib.abc.MetaPathFinder):
    def __init__(self, tracer):
        self._tracer = tracer
        self._disabled = False

    def find_spec(self, fullname, path=None, target=None):
        if self._disabled:
            return None

        if fullname in RECORDER_SPECS:
            try:
                self._disabled = True
                spec = importlib.util.find_spec(fullname)
                if spec:
                    loader = importlib.util.find_spec(fullname).loader
                    if loader is not None and isinstance(loader, importlib.abc.SourceLoader):
                        return importlib.util.spec_from_loader(fullname, SourceLoaderWrapper(loader, self._tracer))
            except Exception:
                logger.error('Error patching spec for module %s', fullname, exc_info=True)
            finally:
                self._disabled = False

        return None


class Tracer:
    METRIC_READ_INTERVAL_SEC = 10
    METRIC_UPLOAD_INTERVAL_SEC = 20
    LOG_UPLOAD_INTERVAL_SEC = 20

    def __init__(
            self, 
            api_key=None, 
            api_url=None, 
            deployment=None, 
            tags=None, 
            auto_instrument=True, 
            record_data_samples=False, 
            upload_on_shutdown=True, 
            debug_mode=False):
        if debug_mode:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.WARNING)

        self.api_key = api_key
        if api_url:
            self.api_url = api_url
        else:
            self.api_url = 'https://signal-api.graphsignal.com'
        self.deployment = deployment
        self.tags = tags
        self.context_tags = None
        self.auto_instrument = auto_instrument
        self.record_data_samples = record_data_samples
        self.upload_on_shutdown = upload_on_shutdown
        self.debug_mode = debug_mode
        self.hostname = None
        try:
            self.hostname = socket.gethostname()
        except BaseException:
            logger.debug('Error reading hostname', exc_info=True)

        self._tracer_log_handler = None
        self._uploader = None
        self._metric_store = None
        self._log_store = None
        self._recorders = None
        self._random_samplers = None
        self._lo_samplers = None
        self._mv_samplers = None

        self._process_start_ms = int(time.time() * 1e3)

        self.last_metric_read_ts = 0
        self.last_metric_upload_ts = int(self._process_start_ms / 1e3)
        self.last_log_upload_ts = int(self._process_start_ms / 1e3)

    def setup(self):
        self._uploader = Uploader()
        self._uploader.setup()
        self._metric_store = MetricStore()
        self._log_store = LogStore()
        self._recorders = {}
        self._random_samplers = {}
        self._lo_samplers = {}
        self._mv_samplers = {}

        # initialize tracer log handler
        self._tracer_log_handler = GraphsignalTracerLogHandler(self)
        logger.addHandler(self._tracer_log_handler)

        # initialize context tags variable
        self.context_tags = contextvars.ContextVar('graphsignal_context_tags', default={})

        # initialize module wrappers
        self._module_finder = SupportedModuleFinder(self)
        sys.meta_path.insert(0, self._module_finder)

        # initialize default recorders and recorders for already loaded modules
        for module_name in RECORDER_SPECS.keys():
            if module_name == '(default)' or module_name in sys.modules:
                self.initialize_recorders_for_module(module_name)

    def shutdown(self):
        if self.upload_on_shutdown:
            if self._metric_store.has_unexported():
                metrics = self._metric_store.export()
                for metric in metrics:
                    self._uploader.upload_metric(metric)

            if self._log_store.has_unexported():
                entries = self._log_store.export()
                for entry in entries:
                    self._uploader.upload_log_entry(entry)

        for recorder in self.recorders():
            recorder.shutdown()

        self.upload(block=True)

        self._recorders = None
        self._metric_store = None
        self._log_store = None
        self._lo_samplers = None
        self._random_samplers = None
        self._mv_samplers = None
        self._uploader = None

        self.context_tags.set({})
        self.context_tags = None

        # remove module wrappers
        if self._module_finder in sys.meta_path:
            sys.meta_path.remove(self._module_finder)
        self._module_finder = None

        # remove tracer log handler
        logger.removeHandler(self._tracer_log_handler)
        self._tracer_log_handler = None

    def uploader(self):
        return self._uploader

    def initialize_recorders_for_module(self, module_name):
        # check already loaded
        if module_name in self._recorders:
            return

        # check if supported
        if module_name not in RECORDER_SPECS:
            return

        # load recorder
        logger.debug('Initializing recorder for module: %s', module_name)
        specs = RECORDER_SPECS[module_name]
        self._recorders[module_name] = []
        for spec in specs:
            try:
                recorder_module = importlib.import_module(spec[0])
                recorder_class = getattr(recorder_module, spec[1])
                recorder = recorder_class()
                recorder.setup()
                self._recorders[module_name].append(recorder)
            except Exception:
                logger.error('Failed to initialize recorder for module: %s', module_name, exc_info=True)

    def recorders(self):
        for recorder_list in self._recorders.values():
            for recorder in recorder_list:
                yield recorder

    def metric_store(self):
        return self._metric_store

    def log_store(self):
        return self._log_store

    def random_sampler(self, operation):
        if operation in self._random_samplers:
            return self._random_samplers[operation]
        else:
            random_sampler = self._random_samplers[operation] = RandomSampler()
            return random_sampler

    def lo_sampler(self, operation):
        if operation in self._lo_samplers:
            return self._lo_samplers[operation]
        else:
            lo_sampler = self._lo_samplers[operation] = LatencyOutlierSampler()
            return lo_sampler

    def mv_sampler(self, operation):
        if operation in self._mv_samplers:
            return self._mv_samplers[operation]
        else:
            lo_sampler = self._mv_samplers[operation] = MissingValueSampler()
            return lo_sampler

    def emit_span_start(self, proto, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_span_start(proto, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_span_stop(self, proto, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_span_stop(proto, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_span_read(self, proto, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_span_read(proto, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_metric_update(self):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_metric_update()
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def upload(self, block=False):
        if block:
            self._uploader.flush()
        else:
            self._uploader.flush_in_thread()

    def check_metric_read_interval(self, now=None):
        if now is None:
            now = time.time()
        return (self.last_metric_read_ts < now - Tracer.METRIC_READ_INTERVAL_SEC)
    
    def set_metric_read(self, now=None):
        self.last_metric_read_ts = now if now else time.time()

    def check_metric_upload_interval(self, now=None):
        if now is None:
            now = time.time()
        return (self.last_metric_upload_ts < now - Tracer.METRIC_UPLOAD_INTERVAL_SEC)

    def set_metric_upload(self, now=None):
        self.last_metric_upload_ts = now if now else time.time()

    def check_log_upload_interval(self, now=None):
        if now is None:
            now = time.time()
        return (self.last_log_upload_ts < now - Tracer.LOG_UPLOAD_INTERVAL_SEC)

    def set_log_upload(self, now=None):
        self.last_log_upload_ts = now if now else time.time()

    def tick(self, block=False, now=None):
        if now is None:
            now = time.time()

        if self.check_metric_upload_interval(now):
            if self._metric_store.has_unexported():
                metrics = self._metric_store.export()
                for metric in metrics:
                    self.uploader().upload_metric(metric)
                self.set_metric_upload(now)

        if self.check_log_upload_interval(now):
            if self._log_store.has_unexported():
                entrys = self._log_store.export()
                for entry in entrys:
                    self.uploader().upload_log_entry(entry)
                self.set_log_upload(now)

        self.upload(block=False)



import logging
import sys
import json

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.base_data_profiler import BaseDataProfiler, DataStats, DataSample

logger = logging.getLogger('graphsignal')


class BuiltInTypesProfiler(BaseDataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        return True

    def compute_stats(self, data):
        shape = None
        counts = {}
        if isinstance(data, bytes):
            counts['byte_count'] = len(data)
        elif isinstance(data, str):
            counts['char_count'] = len(data)
            if data == '':
                counts['empty_count'] = 1
        elif isinstance(data, (list, tuple, set)):
            if _is_array(data):
                shape = _shape(data)
                counts['element_count'] = _elems(data)
            else:
                counts['element_count'] = len(data)
        elif data is None:
            counts['null_count'] = 1

        return DataStats(type_name=type(data).__name__, shape=shape, counts=counts)

    def encode_sample(self, data):
        data_dict = _obj_to_dict(data)
        return DataSample(
            content_type='application/json', 
            content_bytes=json.dumps(data_dict).encode('utf-8'))


def _is_array(data):
    if isinstance(data, list):
        if len(data) > 0:
            if isinstance(data[0], (list, str, bytes, int, float, complex)):
                return True
        else:
            return True

    return False


def _shape(data):
    if isinstance(data, list):
        if len(data) == 0:
            return [0]
        else:
            return [len(data)] + _shape(data[0])
    return []


def _elems(data):
    if isinstance(data, list):
         return sum([_elems(elem) for elem in data])
    else:
        return 1


def _obj_to_dict(obj, level=0):
    if level >= 10:
        return
    if isinstance(obj, (int, float, str, bool, type(None))):
        return obj
    elif isinstance(obj, dict):
        return {k: _obj_to_dict(v, level=level+1) for k, v in obj.items()}
    elif isinstance(obj, (list, set, tuple)):
        return [_obj_to_dict(e, level=level+1) for e in obj]
    elif hasattr(obj, '__dict__'):
        return _obj_to_dict(vars(obj), level=level+1)
    else:
        return str(obj)